@@ -2,6 +2,7 @@
 """Tests for the pixel_data_handlers.util module."""
 
 import os
+import random
 from struct import unpack, pack
 from sys import byteorder
 
@@ -13,8 +14,8 @@ try:
 except ImportError:
     HAVE_NP = False
 
-from pydicom import dcmread
-from pydicom.data import get_testdata_files, get_palette_files
+from pydicom import dcmread, config
+from pydicom.data import get_testdata_file, get_palette_files
 from pydicom.dataset import Dataset, FileMetaDataset
 from pydicom.pixel_data_handlers.util import (
     dtype_corrected_for_endianness,
@@ -26,6 +27,10 @@ from pydicom.pixel_data_handlers.util import (
     _expand_segmented_lut,
     apply_modality_lut,
     apply_voi_lut,
+    get_j2k_parameters,
+    get_nr_frames,
+    apply_voi,
+    apply_windowing
 )
 from pydicom.uid import (ExplicitVRLittleEndian, ImplicitVRLittleEndian,
                          UncompressedPixelTransferSyntaxes)
@@ -36,27 +41,27 @@ from pydicom.uid import (ExplicitVRLittleEndian, ImplicitVRLittleEndian,
 # SUP: Supplemental Palette Color
 # LE, BE: little endian, big endian encoding
 # 8/8, 1 sample/pixel, 1 frame
-PAL_08_256_0_16_1F = get_testdata_files("OBXXXX1A.dcm")[0]
-PAL_08_200_0_16_1F = get_testdata_files("OT-PAL-8-face.dcm")[0]
+PAL_08_256_0_16_1F = get_testdata_file("OBXXXX1A.dcm")
+PAL_08_200_0_16_1F = get_testdata_file("OT-PAL-8-face.dcm")
 # 8/8, 1 sample/pixel, 2 frame
-PAL_08_256_0_16_2F = get_testdata_files("OBXXXX1A_2frame.dcm")[0]
+PAL_08_256_0_16_2F = get_testdata_file("OBXXXX1A_2frame.dcm")
 # PALETTE COLOR with 16-bit LUTs (no indirect segments)
-PAL_SEG_LE_16_1F = get_testdata_files("gdcm-US-ALOKA-16.dcm")[0]
-PAL_SEG_BE_16_1F = get_testdata_files("gdcm-US-ALOKA-16_big.dcm")[0]
+PAL_SEG_LE_16_1F = get_testdata_file("gdcm-US-ALOKA-16.dcm")
+PAL_SEG_BE_16_1F = get_testdata_file("gdcm-US-ALOKA-16_big.dcm")
 # Supplemental palette colour + VOI windowing
-SUP_16_16_2F = get_testdata_files("eCT_Supplemental.dcm")[0]
+SUP_16_16_2F = get_testdata_file("eCT_Supplemental.dcm")
 # 8 bit, 3 samples/pixel, 1 and 2 frame datasets
 # RGB colorspace, uncompressed
-RGB_8_3_1F = get_testdata_files("SC_rgb.dcm")[0]
-RGB_8_3_2F = get_testdata_files("SC_rgb_2frame.dcm")[0]
+RGB_8_3_1F = get_testdata_file("SC_rgb.dcm")
+RGB_8_3_2F = get_testdata_file("SC_rgb_2frame.dcm")
 # MOD: Modality LUT
 # SEQ: Modality LUT Sequence
-MOD_16 = get_testdata_files("CT_small.dcm")[0]
-MOD_16_SEQ = get_testdata_files("mlut_18.dcm")[0]
+MOD_16 = get_testdata_file("CT_small.dcm")
+MOD_16_SEQ = get_testdata_file("mlut_18.dcm")
 # VOI: VOI LUT Sequence
 # WIN: Windowing operation
-WIN_12_1F = get_testdata_files("MR-SIEMENS-DICOM-WithOverlays.dcm")[0]
-VOI_08_1F = get_testdata_files("vlut_04.dcm")[0]
+WIN_12_1F = get_testdata_file("MR-SIEMENS-DICOM-WithOverlays.dcm")
+VOI_08_1F = get_testdata_file("vlut_04.dcm")
 
 
 # Tests with Numpy unavailable
@@ -467,16 +472,15 @@ class TestNumpy_ReshapePixelArray:
 
     def test_compressed_syntaxes_1conf(self):
         """Test the compressed syntaxes that are always 1 planar conf."""
-        for uid in ['1.2.840.10008.1.2.4.80',
-                    '1.2.840.10008.1.2.4.81',
-                    '1.2.840.10008.1.2.5']:
+        for uid in ['1.2.840.10008.1.2.5']:
             self.ds.file_meta.TransferSyntaxUID = uid
             self.ds.PlanarConfiguration = 0
             self.ds.NumberOfFrames = 1
             self.ds.SamplesPerPixel = 3
 
-            arr = reshape_pixel_array(self.ds,
-                                      RESHAPE_ARRAYS['1frame_3sample_1config'])
+            arr = reshape_pixel_array(
+                self.ds, RESHAPE_ARRAYS['1frame_3sample_1config']
+            )
             assert (4, 5, 3) == arr.shape
             assert np.array_equal(arr, self.ref_1_3)
 
@@ -633,6 +637,34 @@ class TestNumpy_ConvertColourSpace:
         assert np.allclose(rgb, arr, atol=1)
         assert rgb.shape == arr.shape
 
+    def test_frame_by_frame(self):
+        """Test processing frame-by-frame."""
+        ds = dcmread(RGB_8_3_2F)
+
+        arr = ds.pixel_array
+        ybr = convert_color_space(arr, 'RGB', 'YBR_FULL', per_frame=True)
+        assert (76, 85, 255) == tuple(ybr[0, 5, 50, :])
+        assert (166, 107, 192) == tuple(ybr[0, 15, 50, :])
+        assert (150, 44, 21) == tuple(ybr[0, 25, 50, :])
+        assert (203, 86, 75) == tuple(ybr[0, 35, 50, :])
+        assert (29, 255, 107) == tuple(ybr[0, 45, 50, :])
+        assert (142, 192, 118) == tuple(ybr[0, 55, 50, :])
+        assert (0, 128, 128) == tuple(ybr[0, 65, 50, :])
+        assert (64, 128, 128) == tuple(ybr[0, 75, 50, :])
+        assert (192, 128, 128) == tuple(ybr[0, 85, 50, :])
+        assert (255, 128, 128) == tuple(ybr[0, 95, 50, :])
+        # Frame 2
+        assert (179, 171, 1) == tuple(ybr[1, 5, 50, :])
+        assert (89, 149, 65) == tuple(ybr[1, 15, 50, :])
+        assert (105, 212, 235) == tuple(ybr[1, 25, 50, :])
+        assert (52, 170, 181) == tuple(ybr[1, 35, 50, :])
+        assert (226, 1, 149) == tuple(ybr[1, 45, 50, :])
+        assert (113, 65, 138) == tuple(ybr[1, 55, 50, :])
+        assert (255, 128, 128) == tuple(ybr[1, 65, 50, :])
+        assert (191, 128, 128) == tuple(ybr[1, 75, 50, :])
+        assert (63, 128, 128) == tuple(ybr[1, 85, 50, :])
+        assert (0, 128, 128) == tuple(ybr[1, 95, 50, :])
+
 
 @pytest.mark.skipif(not HAVE_NP, reason="Numpy is not available")
 class TestNumpy_DtypeCorrectedForEndianness:
@@ -944,9 +976,7 @@ class TestNumpy_PaletteColor:
         """Test that an invalid bit depth raises an exception."""
         ds = dcmread(PAL_08_256_0_16_1F)
         ds.RedPaletteColorLookupTableDescriptor[2] = 15
-        msg = (
-            r'data type "uint15" not understood'
-        )
+        msg = r"data type ['\"]uint15['\"] not understood"
         with pytest.raises(TypeError, match=msg):
             apply_color_lut(ds.pixel_array, ds)
 
@@ -1222,6 +1252,11 @@ class TestNumpy_ExpandSegmentedLUT:
         out = _expand_segmented_lut(data, 'H')
         assert [-400, -320, -240, -160, -80, 0] == out
 
+        # Positive slope, floating point steps
+        data = (0, 1, 163, 1, 48, 255)
+        out = _expand_segmented_lut(data, 'H')
+        assert (1 + 48) == len(out)
+
         # No slope
         data = (0, 2, 0, 28672, 1, 5, 28672)
         out = _expand_segmented_lut(data, 'H')
@@ -1412,172 +1447,8 @@ class TestNumpy_ExpandSegmentedLUT:
 
 
 @pytest.mark.skipif(not HAVE_NP, reason="Numpy is not available")
-class TestNumpy_VOILUT:
-    """Tests for util.apply_voi_lut()."""
-    def test_voi_single_view(self):
-        """Test VOI LUT with a single view."""
-        ds = dcmread(VOI_08_1F)
-        assert 8 == ds.BitsAllocated
-        assert 8 == ds.BitsStored
-        assert 0 == ds.PixelRepresentation
-        item = ds.VOILUTSequence[0]
-        assert [256, 0, 16] == item.LUTDescriptor
-        lut = item.LUTData
-        assert 0 == lut[0]
-        assert 19532 == lut[76]
-        assert 45746 == lut[178]
-        assert 65535 == lut[255]
-
-        arr = ds.pixel_array
-        assert 0 == arr[387, 448]
-        assert 76 == arr[178, 126]
-        assert 178 == arr[186, 389]
-        assert 255 == arr[129, 79]
-
-        out = apply_voi_lut(arr, ds)
-        assert 0 == out[387, 448]
-        assert 19532 == out[178, 126]
-        assert 45746 == out[186, 389]
-        assert 65535 == out[129, 79]
-
-    def test_voi_multi_view(self):
-        """Test VOI LUT with multiple views."""
-        ds = dcmread(VOI_08_1F)
-        assert 8 == ds.BitsAllocated
-        assert 8 == ds.BitsStored
-        assert 0 == ds.PixelRepresentation
-        item0 = ds.VOILUTSequence[0]
-        # Add another view thats the inverse
-        ds.VOILUTSequence.append(Dataset())
-        item1 = ds.VOILUTSequence[1]
-        item1.LUTDescriptor = [256, 0, 16]
-        item1.LUTData = item0.LUTData[::-1]
-
-        arr = ds.pixel_array
-        assert 0 == arr[387, 448]
-        assert 76 == arr[178, 126]
-        assert 178 == arr[186, 389]
-        assert 255 == arr[129, 79]
-
-        out0 = apply_voi_lut(arr, ds)
-        assert 0 == out0[387, 448]
-        assert 19532 == out0[178, 126]
-        assert 45746 == out0[186, 389]
-        assert 65535 == out0[129, 79]
-
-        out1 = apply_voi_lut(arr, ds, index=1)
-        assert 65535 == out1[387, 448]
-        assert 46003 == out1[178, 126]
-        assert 19789 == out1[186, 389]
-        assert 0 == out1[129, 79]
-
-    def test_voi_multi_frame(self):
-        """Test VOI with a multiple frames."""
-        ds = dcmread(VOI_08_1F)
-        assert 8 == ds.BitsAllocated
-        assert 8 == ds.BitsStored
-        assert 0 == ds.PixelRepresentation
-
-        arr = ds.pixel_array
-        arr = np.stack([arr, 255 - arr])
-        assert (2, 512, 512) == arr.shape
-
-        out = apply_voi_lut(arr, ds)
-        assert 0 == out[0, 387, 448]
-        assert 19532 == out[0, 178, 126]
-        assert 45746 == out[0, 186, 389]
-        assert 65535 == out[0, 129, 79]
-        assert 65535 == out[1, 387, 448]
-        assert 46003 == out[1, 178, 126]
-        assert 19789 == out[1, 186, 389]
-        assert 0 == out[1, 129, 79]
-
-    def test_voi_zero_entries(self):
-        """Test that 0 entries is interpreted correctly."""
-        ds = dcmread(VOI_08_1F)
-        seq = ds.VOILUTSequence[0]
-        seq.LUTDescriptor = [0, 0, 16]
-        assert 256 == len(seq.LUTData)
-        arr = np.asarray([0, 255, 256, 65535])
-        msg = r"index 256 is out of bounds"
-        with pytest.raises(IndexError, match=msg):
-            apply_voi_lut(arr, ds)
-
-        # LUTData with 65536 entries
-        seq.LUTData = [0] * 65535 + [1]
-        out = apply_voi_lut(arr, ds)
-        assert [0, 0, 0, 1] == list(out)
-
-    def test_voi_uint8(self):
-        """Test uint VOI LUT with an 8-bit LUT."""
-        ds = Dataset()
-        ds.PixelRepresentation = 0
-        ds.BitsStored = 8
-        ds.VOILUTSequence = [Dataset()]
-        item = ds.VOILUTSequence[0]
-        item.LUTDescriptor = [4, 0, 8]
-        item.LUTData = [0, 127, 128, 255]
-        arr = np.asarray([0, 1, 128, 254, 255], dtype='uint8')
-        out = apply_voi_lut(arr, ds)
-        assert 'uint8' == out.dtype
-        assert [0, 127, 255, 255, 255] == out.tolist()
-
-    def test_voi_uint16(self):
-        """Test uint VOI LUT with an 16-bit LUT."""
-        ds = Dataset()
-        ds.PixelRepresentation = 0
-        ds.BitsStored = 16
-        ds.VOILUTSequence = [Dataset()]
-        item = ds.VOILUTSequence[0]
-        item.LUTDescriptor = [4, 0, 16]
-        item.LUTData = [0, 127, 32768, 65535]
-        arr = np.asarray([0, 1, 2, 3, 255], dtype='uint16')
-        out = apply_voi_lut(arr, ds)
-        assert 'uint16' == out.dtype
-        assert [0, 127, 32768, 65535, 65535] == out.tolist()
-
-    def test_voi_int8(self):
-        """Test int VOI LUT with an 8-bit LUT."""
-        ds = Dataset()
-        ds.PixelRepresentation = 1
-        ds.BitsStored = 8
-        ds.VOILUTSequence = [Dataset()]
-        item = ds.VOILUTSequence[0]
-        item.LUTDescriptor = [4, 0, 8]
-        item.LUTData = [0, 127, 128, 255]
-        arr = np.asarray([0, -1, 2, -128, 127], dtype='int8')
-        out = apply_voi_lut(arr, ds)
-        assert 'uint8' == out.dtype
-        assert [0, 0, 128, 0, 255] == out.tolist()
-
-    def test_voi_int16(self):
-        """Test int VOI LUT with an 16-bit LUT."""
-        ds = Dataset()
-        ds.PixelRepresentation = 0
-        ds.BitsStored = 16
-        ds.VOILUTSequence = [Dataset()]
-        item = ds.VOILUTSequence[0]
-        item.LUTDescriptor = [4, 0, 16]
-        item.LUTData = [0, 127, 32768, 65535]
-        arr = np.asarray([0, -1, 2, -128, 255], dtype='int16')
-        out = apply_voi_lut(arr, ds)
-        assert 'uint16' == out.dtype
-        assert [0, 0, 32768, 0, 65535] == out.tolist()
-
-    def test_voi_bad_depth(self):
-        """Test bad LUT depth raises exception."""
-        ds = dcmread(VOI_08_1F)
-        item = ds.VOILUTSequence[0]
-        item.LUTDescriptor[2] = 7
-        msg = r"'7' bits per LUT entry is not supported"
-        with pytest.raises(NotImplementedError, match=msg):
-            apply_voi_lut(ds.pixel_array, ds)
-
-        item.LUTDescriptor[2] = 17
-        msg = r"'17' bits per LUT entry is not supported"
-        with pytest.raises(NotImplementedError, match=msg):
-            apply_voi_lut(ds.pixel_array, ds)
-
+class TestNumpy_ApplyWindowing:
+    """Tests for util.apply_windowing()."""
     def test_window_single_view(self):
         """Test windowing with a single view."""
         # 12-bit unsigned
@@ -1592,7 +1463,7 @@ class TestNumpy_VOILUT:
 
         arr = ds.pixel_array
         assert 642 == arr[326, 130]
-        out = apply_voi_lut(arr, ds)
+        out = apply_windowing(arr, ds)
         assert 3046.6 == pytest.approx(out[326, 130], abs=0.1)
 
     def test_window_multi_view(self):
@@ -1601,14 +1472,20 @@ class TestNumpy_VOILUT:
         assert 16 == ds.BitsAllocated
         assert 12 == ds.BitsStored
         assert 0 == ds.PixelRepresentation
-        assert [450, 200] == ds.WindowCenter
-        assert [790, 443] == ds.WindowWidth
+        if HAVE_NP and config.use_DS_numpy:
+            expected = np.array([450, 200])
+            assert np.allclose(ds.WindowCenter, expected)
+            expected = np.array([790, 443])
+            assert np.allclose(ds.WindowWidth, expected)
+        else:
+            assert [450, 200] == ds.WindowCenter
+            assert [790, 443] == ds.WindowWidth
 
         arr = ds.pixel_array
         assert 642 == arr[326, 130]
-        out = apply_voi_lut(arr, ds)
+        out = apply_windowing(arr, ds)
         assert 3046.6 == pytest.approx(out[326, 130], abs=0.1)
-        out = apply_voi_lut(arr, ds, index=1)
+        out = apply_windowing(arr, ds, index=1)
         assert 4095.0 == pytest.approx(out[326, 130], abs=0.1)
 
     def test_window_uint8(self):
@@ -1622,24 +1499,24 @@ class TestNumpy_VOILUT:
         # Linear
         ds.WindowWidth = 1
         ds.WindowCenter = 0
-        assert [255, 255, 255, 255, 255] == apply_voi_lut(arr, ds).tolist()
+        assert [255, 255, 255, 255, 255] == apply_windowing(arr, ds).tolist()
 
         ds.WindowWidth = 128
         ds.WindowCenter = 254
         assert [0, 0, 0, 128.5, 130.5] == pytest.approx(
-            apply_voi_lut(arr, ds).tolist(), abs=0.1
+            apply_windowing(arr, ds).tolist(), abs=0.1
         )
 
         # Linear exact
         ds.VOILUTFunction = 'LINEAR_EXACT'
         assert [0, 0, 0, 127.5, 129.5] == pytest.approx(
-            apply_voi_lut(arr, ds).tolist(), abs=0.1
+            apply_windowing(arr, ds).tolist(), abs=0.1
         )
 
         # Sigmoid
         ds.VOILUTFunction = 'SIGMOID'
         assert [0.1, 0.1, 4.9, 127.5, 129.5] == pytest.approx(
-            apply_voi_lut(arr, ds).tolist(), abs=0.1
+            apply_windowing(arr, ds).tolist(), abs=0.1
         )
 
     def test_window_uint16(self):
@@ -1652,22 +1529,22 @@ class TestNumpy_VOILUT:
 
         ds.WindowWidth = 1
         ds.WindowCenter = 0
-        assert [65535] * 5 == apply_voi_lut(arr, ds).tolist()
+        assert [65535] * 5 == apply_windowing(arr, ds).tolist()
 
         ds.WindowWidth = 32768
         ds.WindowCenter = 254
         assert [32260.5, 32262.5, 65535, 65535, 65535] == pytest.approx(
-            apply_voi_lut(arr, ds).tolist(), abs=0.1
+            apply_windowing(arr, ds).tolist(), abs=0.1
         )
 
         ds.VOILUTFunction = 'LINEAR_EXACT'
         assert [32259.5, 32261.5, 65535, 65535, 65535] == pytest.approx(
-            apply_voi_lut(arr, ds).tolist(), abs=0.1
+            apply_windowing(arr, ds).tolist(), abs=0.1
         )
 
         ds.VOILUTFunction = 'SIGMOID'
         assert [32259.5, 32261.5, 64319.8, 65512.3, 65512.3] == pytest.approx(
-            apply_voi_lut(arr, ds).tolist(), abs=0.1
+            apply_windowing(arr, ds).tolist(), abs=0.1
         )
 
     def test_window_uint32(self):
@@ -1681,14 +1558,14 @@ class TestNumpy_VOILUT:
 
         ds.WindowWidth = 1
         ds.WindowCenter = 0
-        assert [y_max] * 5 == apply_voi_lut(arr, ds).tolist()
+        assert [y_max] * 5 == apply_windowing(arr, ds).tolist()
 
         ds.WindowWidth = 342423423423
         ds.WindowCenter = 757336
         assert (
             [2147474148.4, 2147474148.4,
              2174409724, 2201345299.7, 2201345299.7] == pytest.approx(
-                apply_voi_lut(arr, ds).tolist(), abs=0.1
+                apply_windowing(arr, ds).tolist(), abs=0.1
             )
         )
 
@@ -1696,7 +1573,7 @@ class TestNumpy_VOILUT:
         assert (
             [2147474148.3, 2147474148.4,
              2174409724, 2201345299.7, 2201345299.7] == pytest.approx(
-                apply_voi_lut(arr, ds).tolist(), abs=0.1
+                apply_windowing(arr, ds).tolist(), abs=0.1
             )
         )
 
@@ -1704,7 +1581,7 @@ class TestNumpy_VOILUT:
         assert (
             [2147474148.3, 2147474148.4,
              2174408313.1, 2201334008.2, 2201334008.3] == pytest.approx(
-                apply_voi_lut(arr, ds).tolist(), abs=0.1
+                apply_windowing(arr, ds).tolist(), abs=0.1
             )
         )
 
@@ -1720,25 +1597,25 @@ class TestNumpy_VOILUT:
         ds.WindowWidth = 1
         ds.WindowCenter = 0
         assert [-128, -128, -128, 127, 127, 127, 127] == pytest.approx(
-            apply_voi_lut(arr, ds).tolist()
+            apply_windowing(arr, ds).tolist()
         )
 
         ds.WindowWidth = 128
         ds.WindowCenter = -5
         assert [-128, -128, 8.5, 10.5, 12.6, 127, 127] == pytest.approx(
-            apply_voi_lut(arr, ds).tolist(), abs=0.1
+            apply_windowing(arr, ds).tolist(), abs=0.1
         )
 
         # Linear exact
         ds.VOILUTFunction = 'LINEAR_EXACT'
         assert [-128, -128, 7.5, 9.5, 11.5, 127, 127] == pytest.approx(
-            apply_voi_lut(arr, ds).tolist(), abs=0.1
+            apply_windowing(arr, ds).tolist(), abs=0.1
         )
 
         # Sigmoid
         ds.VOILUTFunction = 'SIGMOID'
         assert [-122.7, -122.5, 7.5, 9.4, 11.4, 122.8, 122.9] == pytest.approx(
-            apply_voi_lut(arr, ds).tolist(), abs=0.1
+            apply_windowing(arr, ds).tolist(), abs=0.1
         )
 
     def test_window_int16(self):
@@ -1755,7 +1632,7 @@ class TestNumpy_VOILUT:
         assert (
             [-32768, -32768, -32768,
              32767, 32767, 32767, 32767] == pytest.approx(
-                apply_voi_lut(arr, ds).tolist(), abs=0.1
+                apply_windowing(arr, ds).tolist(), abs=0.1
             )
         )
 
@@ -1764,7 +1641,7 @@ class TestNumpy_VOILUT:
         assert (
             [-32768, -32768, 2321.6,
              2837.6, 3353.7, 32767, 32767] == pytest.approx(
-                apply_voi_lut(arr, ds).tolist(), abs=0.1
+                apply_windowing(arr, ds).tolist(), abs=0.1
             )
         )
 
@@ -1773,7 +1650,7 @@ class TestNumpy_VOILUT:
         assert (
             [-32768, -32768, 2047.5,
              2559.5, 3071.5, 32767, 32767] == pytest.approx(
-                apply_voi_lut(arr, ds).tolist(), abs=0.1
+                apply_windowing(arr, ds).tolist(), abs=0.1
             )
         )
 
@@ -1782,7 +1659,7 @@ class TestNumpy_VOILUT:
         assert (
             [-31394.1, -31351.4, 2044.8,
              2554.3, 3062.5, 31692, 31724.6] == pytest.approx(
-                apply_voi_lut(arr, ds).tolist(), abs=0.1
+                apply_windowing(arr, ds).tolist(), abs=0.1
             )
         )
 
@@ -1800,7 +1677,7 @@ class TestNumpy_VOILUT:
         assert (
             [-2**31, -2**31, -2**31,
              2**31 - 1, 2**31 - 1, 2**31 - 1, 2**31 - 1] == pytest.approx(
-                apply_voi_lut(arr, ds).tolist(), abs=0.1
+                apply_windowing(arr, ds).tolist(), abs=0.1
             )
         )
 
@@ -1809,7 +1686,7 @@ class TestNumpy_VOILUT:
         assert (
             [-2147483648, -2147483648, 152183880, 186002520.1,
              219821160.3, 2147483647, 2147483647] == pytest.approx(
-                apply_voi_lut(arr, ds).tolist(), abs=0.1
+                apply_windowing(arr, ds).tolist(), abs=0.1
             )
         )
 
@@ -1818,7 +1695,7 @@ class TestNumpy_VOILUT:
         assert (
             [-2147483648, -2147483648, 134217727.5, 167772159.5,
              201326591.5, 2147483647, 2147483647] == pytest.approx(
-                apply_voi_lut(arr, ds).tolist(), abs=0.1
+                apply_windowing(arr, ds).tolist(), abs=0.1
             )
         )
 
@@ -1827,7 +1704,7 @@ class TestNumpy_VOILUT:
         assert (
             [-2057442919.3, -2054646500.7, 134043237.4, 167431657.4,
              200738833.7, 2077033158.8, 2079166214.8] == pytest.approx(
-                apply_voi_lut(arr, ds).tolist(), abs=0.1
+                apply_windowing(arr, ds).tolist(), abs=0.1
             )
         )
 
@@ -1847,7 +1724,7 @@ class TestNumpy_VOILUT:
         assert (2, 484, 484) == arr.shape
         assert 642 == arr[0, 326, 130]
         assert 3453 == arr[1, 326, 130]
-        out = apply_voi_lut(arr, ds)
+        out = apply_windowing(arr, ds)
         assert 3046.6 == pytest.approx(out[0, 326, 130], abs=0.1)
         assert 4095.0 == pytest.approx(out[1, 326, 130], abs=0.1)
 
@@ -1857,8 +1734,14 @@ class TestNumpy_VOILUT:
         assert 16 == ds.BitsAllocated
         assert 12 == ds.BitsStored
         assert 0 == ds.PixelRepresentation
-        assert [450, 200] == ds.WindowCenter
-        assert [790, 443] == ds.WindowWidth
+        if HAVE_NP and config.use_DS_numpy:
+            expected = np.array([450, 200])
+            assert np.allclose(ds.WindowCenter, expected)
+            expected = np.array([790, 443])
+            assert np.allclose(ds.WindowWidth, expected)
+        else:
+            assert [450, 200] == ds.WindowCenter
+            assert [790, 443] == ds.WindowWidth
         ds.RescaleSlope = 1.2
         ds.RescaleIntercept = 0
 
@@ -1871,7 +1754,7 @@ class TestNumpy_VOILUT:
         assert 770.4 == hu[326, 130]
         assert 1347.6 == hu[316, 481]
         # With rescale -> output range is 0 to 4914
-        out = apply_voi_lut(hu, ds)
+        out = apply_windowing(hu, ds)
         assert 0 == pytest.approx(out[16, 60], abs=0.1)
         assert 4455.6 == pytest.approx(out[326, 130], abs=0.1)
         assert 4914.0 == pytest.approx(out[316, 481], abs=0.1)
@@ -1899,7 +1782,7 @@ class TestNumpy_VOILUT:
         hu = apply_modality_lut(arr, ds)
         assert 65535 == hu[16, 60]
         assert 49147 == hu[0, 1]
-        out = apply_voi_lut(hu, ds)
+        out = apply_windowing(hu, ds)
         assert 65535.0 == pytest.approx(out[16, 60], abs=0.1)
         assert 32809.0 == pytest.approx(out[0, 1], abs=0.1)
         # Output range must be 0 to 2**16 - 1
@@ -1912,7 +1795,7 @@ class TestNumpy_VOILUT:
         ds.PhotometricInterpretation = 'RGB'
         msg = r"only 'MONOCHROME1' and 'MONOCHROME2' are allowed"
         with pytest.raises(ValueError, match=msg):
-            apply_voi_lut(ds.pixel_array, ds)
+            apply_windowing(ds.pixel_array, ds)
 
     def test_window_bad_parameters(self):
         """Test bad windowing parameters raise exceptions."""
@@ -1921,30 +1804,30 @@ class TestNumpy_VOILUT:
         ds.VOILUTFunction = 'LINEAR'
         msg = r"Width must be greater than or equal to 1"
         with pytest.raises(ValueError, match=msg):
-            apply_voi_lut(ds.pixel_array, ds)
+            apply_windowing(ds.pixel_array, ds)
 
         ds.VOILUTFunction = 'LINEAR_EXACT'
         msg = r"Width must be greater than 0"
         with pytest.raises(ValueError, match=msg):
-            apply_voi_lut(ds.pixel_array, ds)
+            apply_windowing(ds.pixel_array, ds)
 
         ds.VOILUTFunction = 'SIGMOID'
         msg = r"Width must be greater than 0"
         with pytest.raises(ValueError, match=msg):
-            apply_voi_lut(ds.pixel_array, ds)
+            apply_windowing(ds.pixel_array, ds)
 
         ds.VOILUTFunction = 'UNKNOWN'
         msg = r"Unsupported \(0028,1056\) VOI LUT Function value 'UNKNOWN'"
         with pytest.raises(ValueError, match=msg):
-            apply_voi_lut(ds.pixel_array, ds)
+            apply_windowing(ds.pixel_array, ds)
 
-    def test_window_bad_index(self):
+    def test_window_bad_index(self, no_numpy_use):
         """Test windowing with a bad view index."""
         ds = dcmread(WIN_12_1F)
         assert 2 == len(ds.WindowWidth)
         arr = ds.pixel_array
         with pytest.raises(IndexError, match=r"list index out of range"):
-            apply_voi_lut(arr, ds, index=2)
+            apply_windowing(arr, ds, index=2)
 
     def test_unchanged(self):
         """Test input array is unchanged if no VOI LUT"""
@@ -1953,7 +1836,219 @@ class TestNumpy_VOILUT:
         ds.PixelRepresentation = 1
         ds.BitsStored = 8
         arr = np.asarray([-128, -127, -1, 0, 1, 126, 127], dtype='int8')
-        out = apply_voi_lut(arr, ds)
+        out = apply_windowing(arr, ds)
+        assert [-128, -127, -1, 0, 1, 126, 127] == out.tolist()
+
+    def test_rescale_empty(self):
+        """Test RescaleSlope and RescaleIntercept being empty."""
+        ds = dcmread(WIN_12_1F)
+        ds.RescaleSlope = None
+        ds.RescaleIntercept = None
+
+        arr = ds.pixel_array
+        assert 0 == arr[16, 60]
+        assert 642 == arr[326, 130]
+        assert 1123 == arr[316, 481]
+        out = apply_windowing(arr, ds)
+        assert 0 == pytest.approx(out[16, 60], abs=0.1)
+        assert 3046.6 == pytest.approx(out[326, 130], abs=0.1)
+        assert 4095.0 == pytest.approx(out[316, 481], abs=0.1)
+
+
+@pytest.mark.skipif(not HAVE_NP, reason="Numpy is not available")
+class TestNumpy_ApplyVOI:
+    """Tests for util.apply_voi()."""
+    def test_voi_single_view(self):
+        """Test VOI LUT with a single view."""
+        ds = dcmread(VOI_08_1F)
+        assert 8 == ds.BitsAllocated
+        assert 8 == ds.BitsStored
+        assert 0 == ds.PixelRepresentation
+        item = ds.VOILUTSequence[0]
+        assert [256, 0, 16] == item.LUTDescriptor
+        lut = item.LUTData
+        assert 0 == lut[0]
+        assert 19532 == lut[76]
+        assert 45746 == lut[178]
+        assert 65535 == lut[255]
+
+        arr = ds.pixel_array
+        assert 0 == arr[387, 448]
+        assert 76 == arr[178, 126]
+        assert 178 == arr[186, 389]
+        assert 255 == arr[129, 79]
+
+        out = apply_voi(arr, ds)
+        assert 0 == out[387, 448]
+        assert 19532 == out[178, 126]
+        assert 45746 == out[186, 389]
+        assert 65535 == out[129, 79]
+
+    def test_voi_multi_view(self):
+        """Test VOI LUT with multiple views."""
+        ds = dcmread(VOI_08_1F)
+        assert 8 == ds.BitsAllocated
+        assert 8 == ds.BitsStored
+        assert 0 == ds.PixelRepresentation
+        item0 = ds.VOILUTSequence[0]
+        # Add another view thats the inverse
+        ds.VOILUTSequence.append(Dataset())
+        item1 = ds.VOILUTSequence[1]
+        item1.LUTDescriptor = [256, 0, 16]
+        item1.LUTData = item0.LUTData[::-1]
+
+        arr = ds.pixel_array
+        assert 0 == arr[387, 448]
+        assert 76 == arr[178, 126]
+        assert 178 == arr[186, 389]
+        assert 255 == arr[129, 79]
+
+        out0 = apply_voi(arr, ds)
+        assert 0 == out0[387, 448]
+        assert 19532 == out0[178, 126]
+        assert 45746 == out0[186, 389]
+        assert 65535 == out0[129, 79]
+
+        out1 = apply_voi(arr, ds, index=1)
+        assert 65535 == out1[387, 448]
+        assert 46003 == out1[178, 126]
+        assert 19789 == out1[186, 389]
+        assert 0 == out1[129, 79]
+
+    def test_voi_multi_frame(self):
+        """Test VOI with a multiple frames."""
+        ds = dcmread(VOI_08_1F)
+        assert 8 == ds.BitsAllocated
+        assert 8 == ds.BitsStored
+        assert 0 == ds.PixelRepresentation
+
+        arr = ds.pixel_array
+        arr = np.stack([arr, 255 - arr])
+        assert (2, 512, 512) == arr.shape
+
+        out = apply_voi(arr, ds)
+        assert 0 == out[0, 387, 448]
+        assert 19532 == out[0, 178, 126]
+        assert 45746 == out[0, 186, 389]
+        assert 65535 == out[0, 129, 79]
+        assert 65535 == out[1, 387, 448]
+        assert 46003 == out[1, 178, 126]
+        assert 19789 == out[1, 186, 389]
+        assert 0 == out[1, 129, 79]
+
+    def test_voi_zero_entries(self):
+        """Test that 0 entries is interpreted correctly."""
+        ds = dcmread(VOI_08_1F)
+        seq = ds.VOILUTSequence[0]
+        seq.LUTDescriptor = [0, 0, 16]
+        assert 256 == len(seq.LUTData)
+        arr = np.asarray([0, 255, 256, 65535])
+        msg = r"index 256 is out of bounds"
+        with pytest.raises(IndexError, match=msg):
+            apply_voi(arr, ds)
+
+        # LUTData with 65536 entries
+        seq.LUTData = [0] * 65535 + [1]
+        out = apply_voi(arr, ds)
+        assert [0, 0, 0, 1] == list(out)
+
+    def test_voi_uint8(self):
+        """Test uint VOI LUT with an 8-bit LUT."""
+        ds = Dataset()
+        ds.PixelRepresentation = 0
+        ds.BitsStored = 8
+        ds.VOILUTSequence = [Dataset()]
+        item = ds.VOILUTSequence[0]
+        item.LUTDescriptor = [4, 0, 8]
+        item.LUTData = [0, 127, 128, 255]
+        arr = np.asarray([0, 1, 128, 254, 255], dtype='uint8')
+        out = apply_voi(arr, ds)
+        assert 'uint8' == out.dtype
+        assert [0, 127, 255, 255, 255] == out.tolist()
+
+    def test_voi_uint16(self):
+        """Test uint VOI LUT with an 16-bit LUT."""
+        ds = Dataset()
+        ds.PixelRepresentation = 0
+        ds.BitsStored = 16
+        ds.VOILUTSequence = [Dataset()]
+        item = ds.VOILUTSequence[0]
+        item.LUTDescriptor = [4, 0, 16]
+        item.LUTData = [0, 127, 32768, 65535]
+        arr = np.asarray([0, 1, 2, 3, 255], dtype='uint16')
+        out = apply_voi(arr, ds)
+        assert 'uint16' == out.dtype
+        assert [0, 127, 32768, 65535, 65535] == out.tolist()
+
+    def test_voi_int8(self):
+        """Test int VOI LUT with an 8-bit LUT."""
+        ds = Dataset()
+        ds.PixelRepresentation = 1
+        ds.BitsStored = 8
+        ds.VOILUTSequence = [Dataset()]
+        item = ds.VOILUTSequence[0]
+        item.LUTDescriptor = [4, 0, 8]
+        item.LUTData = [0, 127, 128, 255]
+        arr = np.asarray([0, -1, 2, -128, 127], dtype='int8')
+        out = apply_voi(arr, ds)
+        assert 'uint8' == out.dtype
+        assert [0, 0, 128, 0, 255] == out.tolist()
+
+    def test_voi_int16(self):
+        """Test int VOI LUT with an 16-bit LUT."""
+        ds = Dataset()
+        ds.PixelRepresentation = 0
+        ds.BitsStored = 16
+        ds.VOILUTSequence = [Dataset()]
+        item = ds.VOILUTSequence[0]
+        item.LUTDescriptor = [4, 0, 16]
+        item.LUTData = [0, 127, 32768, 65535]
+        arr = np.asarray([0, -1, 2, -128, 255], dtype='int16')
+        out = apply_voi(arr, ds)
+        assert 'uint16' == out.dtype
+        assert [0, 0, 32768, 0, 65535] == out.tolist()
+
+    def test_voi_bad_depth(self):
+        """Test bad LUT depth raises exception."""
+        ds = dcmread(VOI_08_1F)
+        item = ds.VOILUTSequence[0]
+        item.LUTDescriptor[2] = 7
+        msg = r"'7' bits per LUT entry is not supported"
+        with pytest.raises(NotImplementedError, match=msg):
+            apply_voi(ds.pixel_array, ds)
+
+        item.LUTDescriptor[2] = 17
+        msg = r"'17' bits per LUT entry is not supported"
+        with pytest.raises(NotImplementedError, match=msg):
+            apply_voi(ds.pixel_array, ds)
+
+    def test_voi_uint16_array_float(self):
+        """Test warning when array is float and VOI LUT with an 16-bit LUT"""
+        ds = Dataset()
+        ds.PixelRepresentation = 0
+        ds.BitsStored = 16
+        ds.VOILUTSequence = [Dataset()]
+        item = ds.VOILUTSequence[0]
+        item.LUTDescriptor = [4, 0, 16]
+        item.LUTData = [0, 127, 32768, 65535]
+        arr = np.asarray([0, 1, 2, 3, 255], dtype='float64')
+        msg = (
+            r"Applying a VOI LUT on a float input array may give "
+            r"incorrect results"
+        )
+
+        with pytest.warns(UserWarning, match=msg):
+            out = apply_voi(arr, ds)
+            assert [0, 127, 32768, 65535, 65535] == out.tolist()
+
+    def test_unchanged(self):
+        """Test input array is unchanged if no VOI LUT"""
+        ds = Dataset()
+        ds.PhotometricInterpretation = 'MONOCHROME1'
+        ds.PixelRepresentation = 1
+        ds.BitsStored = 8
+        arr = np.asarray([-128, -127, -1, 0, 1, 126, 127], dtype='int8')
+        out = apply_voi(arr, ds)
         assert [-128, -127, -1, 0, 1, 126, 127] == out.tolist()
 
     def test_voi_lutdata_ow(self):
@@ -1970,6 +2065,155 @@ class TestNumpy_VOILUT:
         item.LUTData = pack('<4H', *item.LUTData)
         item['LUTData'].VR = 'OW'
         arr = np.asarray([0, 1, 2, 3, 255], dtype='uint16')
-        out = apply_voi_lut(arr, ds)
+        out = apply_voi(arr, ds)
         assert 'uint16' == out.dtype
         assert [0, 127, 32768, 65535, 65535] == out.tolist()
+
+
+@pytest.mark.skipif(not HAVE_NP, reason="Numpy is not available")
+class TestNumpy_ApplyVOILUT:
+    def test_unchanged(self):
+        """Test input array is unchanged if no VOI LUT"""
+        ds = Dataset()
+        ds.PhotometricInterpretation = 'MONOCHROME1'
+        ds.PixelRepresentation = 1
+        ds.BitsStored = 8
+        arr = np.asarray([-128, -127, -1, 0, 1, 126, 127], dtype='int8')
+        out = apply_voi_lut(arr, ds)
+        assert [-128, -127, -1, 0, 1, 126, 127] == out.tolist()
+
+    def test_only_windowing(self):
+        """Test only windowing operation elements present."""
+        ds = Dataset()
+        ds.PhotometricInterpretation = 'MONOCHROME1'
+        ds.PixelRepresentation = 0
+        ds.BitsStored = 8
+        arr = np.asarray([0, 1, 128, 254, 255], dtype='uint8')
+
+        ds.WindowWidth = 1
+        ds.WindowCenter = 0
+        assert [255, 255, 255, 255, 255] == apply_voi_lut(arr, ds).tolist()
+
+    def test_only_voi(self):
+        """Test only LUT operation elements present."""
+        ds = Dataset()
+        ds.PixelRepresentation = 0
+        ds.BitsStored = 8
+        ds.VOILUTSequence = [Dataset()]
+        item = ds.VOILUTSequence[0]
+        item.LUTDescriptor = [4, 0, 8]
+        item.LUTData = [0, 127, 128, 255]
+        arr = np.asarray([0, 1, 128, 254, 255], dtype='uint8')
+        out = apply_voi_lut(arr, ds)
+        assert 'uint8' == out.dtype
+        assert [0, 127, 255, 255, 255] == out.tolist()
+
+    def test_voi_windowing(self):
+        """Test both LUT and windowing operation elements present."""
+        ds = Dataset()
+        ds.PhotometricInterpretation = 'MONOCHROME1'
+        ds.PixelRepresentation = 0
+        ds.BitsStored = 8
+        ds.WindowWidth = 1
+        ds.WindowCenter = 0
+        ds.VOILUTSequence = [Dataset()]
+        item = ds.VOILUTSequence[0]
+        item.LUTDescriptor = [4, 0, 8]
+        item.LUTData = [0, 127, 128, 255]
+        arr = np.asarray([0, 1, 128, 254, 255], dtype='uint8')
+
+        # Defaults to LUT
+        out = apply_voi_lut(arr, ds)
+        assert [0, 127, 255, 255, 255] == out.tolist()
+
+        out = apply_voi_lut(arr, ds, prefer_lut=False)
+        assert [255, 255, 255, 255, 255] == out.tolist()
+
+    def test_voi_windowing_empty(self):
+        """Test empty VOI elements."""
+        ds = Dataset()
+        ds.PhotometricInterpretation = 'MONOCHROME1'
+        ds.PixelRepresentation = 0
+        ds.BitsStored = 8
+        ds.WindowWidth = 1
+        ds.WindowCenter = 0
+        ds.VOILUTSequence = [Dataset()]
+        item = ds.VOILUTSequence[0]
+        item.LUTDescriptor = [4, 0, 8]
+        item.LUTData = [0, 127, 128, 255]
+        arr = np.asarray([0, 1, 128, 254, 255], dtype='uint8')
+
+        # Test empty VOI elements
+        item.LUTData = None
+        out = apply_voi_lut(arr, ds)
+        assert [255, 255, 255, 255, 255] == out.tolist()
+
+        # Test empty windowing elements
+        ds.WindowWidth = None
+        out = apply_voi_lut(arr, ds)
+        assert [0, 1, 128, 254, 255] == out.tolist()
+
+
+class TestGetJ2KParameters:
+    """Tests for get_j2k_parameters."""
+    def test_precision(self):
+        """Test getting the precision for a JPEG2K bytestream."""
+        base = b'\xff\x4f\xff\x51' + b'\x00' * 38
+        # Signed
+        for ii in range(135, 144):
+            params = get_j2k_parameters(base + bytes([ii]))
+            assert ii - 127 == params["precision"]
+            assert params["is_signed"]
+
+        # Unsigned
+        for ii in range(7, 16):
+            params = get_j2k_parameters(base + bytes([ii]))
+            assert ii + 1 == params["precision"]
+            assert not params["is_signed"]
+
+    def test_not_j2k(self):
+        """Test result when no JPEG2K SOF marker present"""
+        base = b'\xff\x4e\xff\x51' + b'\x00' * 38
+        assert {} == get_j2k_parameters(base + b'\x8F')
+
+    def test_no_siz(self):
+        """Test result when no SIZ box present"""
+        base = b'\xff\x4f\xff\x52' + b'\x00' * 38
+        assert {} == get_j2k_parameters(base + b'\x8F')
+
+    def test_short_bytestream(self):
+        """Test result when no SIZ box present"""
+        assert {} == get_j2k_parameters(b'')
+        assert {} == get_j2k_parameters(b'\xff\x4f\xff\x51' + b'\x00' * 20)
+
+
+class TestGetNrFrames:
+    """Tests for get_nr_frames."""
+    def test_none(self):
+        """Test warning when (0028,0008) 'Number of Frames' has a value of
+            None"""
+        ds = Dataset()
+        ds.NumberOfFrames = None
+        msg = (
+            r"A value of None for \(0028,0008\) 'Number of Frames' is "
+            r"non-conformant. It's recommended that this value be "
+            r"changed to 1"
+        )
+        with pytest.warns(UserWarning, match=msg):
+            assert 1 == get_nr_frames(ds)
+
+    def test_missing(self):
+        """Test return value when (0028,0008) 'Number of Frames' does not
+            exist"""
+        ds = Dataset()
+        with pytest.warns(None) as w:
+            assert 1 == get_nr_frames(ds)
+            assert not w
+
+    def test_existing(self):
+        """Test return value when (0028,0008) 'Number of Frames' exists."""
+        ds = Dataset()
+        ds.NumberOfFrames = random.randint(1, 10)
+        with pytest.warns(None) as w:
+            assert ds.NumberOfFrames == get_nr_frames(ds)
+            assert not w
